# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import logging
import shutil
from six.moves import cPickle
from six import iteritems
import tempfile

from .base import BaseConfigEnricher, BaseConfigEnricherThread
from ..errors import BuilderError, ARouteServerError
from ..ipaddresses import IPAddress
from ..irrdb import ASSet, RSet, AS_SET_Bundle

irrdb_pickle_dir = None
def setup_irrdb_pickle_dir():
    global irrdb_pickle_dir

    if irrdb_pickle_dir:
        return irrdb_pickle_dir

    irrdb_pickle_dir = tempfile.mkdtemp(suffix="_arouteserver")
    atexit.register(clear_irrdb_pickle_dir, target_dir=irrdb_pickle_dir)
    return irrdb_pickle_dir

def clear_irrdb_pickle_dir(target_dir):
    shutil.rmtree(target_dir, ignore_errors=True)

class AS_SET_Bundle_Proxy(AS_SET_Bundle):

    def __init__(self, as_set_names):
        AS_SET_Bundle.__init__(self, as_set_names)

        self.used_by = []

        self.irrdb_pickle_dir = setup_irrdb_pickle_dir()

        self.saved_objects = []

    def get_path(self, objects):
        return "{}/pickle_{}.{}".format(self.irrdb_pickle_dir, self.id, objects)

    def save(self, objects, data):
        if objects in ("asns", "prefixes"):
            with open(self.get_path(objects), "wb") as f:
                cPickle.dump(data, f, cPickle.HIGHEST_PROTOCOL)
            self.saved_objects += [objects]
        else:
            raise ValueError("Unknown objects: {}".format(objects))

    def load(self, objects):
        if objects in self.saved_objects:
            with open(self.get_path(objects), "rb") as f:
                return cPickle.load(f)
        else:
            return []

    @property
    def asns(self):
        return self.load("asns")

    @property
    def prefixes(self):
        return self.load("prefixes")

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "descr": self.descr,
            "used_by": ", ".join(sorted(self.used_by)),
            "asns": self.asns,
            "prefixes": self.prefixes
        }

class IRRDBConfigEnricher_WorkerThread(BaseConfigEnricherThread):

    TARGET_FIELD = None

    def __init__(self, *args, **kwargs):
        BaseConfigEnricherThread.__init__(self, *args, **kwargs)

        self.ip_ver = None
        self.irrdbtools_cfg = None

    def do_task(self, task):
        as_set_bundle = task
        used_by_descr = ", ".join(as_set_bundle.used_by)
        data = self._get_external_data(used_by_descr, as_set_bundle.object_names)
        return data

    def save_data(self, task, data):
        as_set_bundle = task
        as_set_bundle.save(self.TARGET_FIELD, data)

class IRRDBConfigEnricher_WorkerThread_Prefixes(IRRDBConfigEnricher_WorkerThread):

    DESCR = "IRRdb prefixes"
    TARGET_FIELD = "prefixes"

    def _get_external_data(self, used_by_descr, as_set_names):
        errors = False
        ip_versions = [self.ip_ver] if self.ip_ver else [4, 6]
        res = []
        for ip_ver in ip_versions:
            obj = RSet(as_set_names, ip_ver, **self.irrdbtools_cfg)
            try:
                obj.load_data()
                prefixes = obj.prefixes
                if not prefixes:
                    logging.warning("No IPv{} prefixes found in "
                                    "{} for {}".format(
                                        ip_ver, obj.descr, used_by_descr))
                res.extend(prefixes)
            except ARouteServerError as e:
                errors = True
                logging.error(
                    "Error while retrieving IPv{} prefixes "
                    "from {} for {}: {}".format(
                        ip_ver, obj.descr, used_by_descr, str(e)
                    )
                )

        if errors:
            raise BuilderError()

        return res

class IRRDBConfigEnricher_WorkerThread_OriginASNs(IRRDBConfigEnricher_WorkerThread):

    DESCR = "IRRdb origin ASNs"
    TARGET_FIELD = "asns"

    def _get_external_data(self, used_by_descr, as_set_name):
        errors = False
        obj = ASSet(as_set_name, **self.irrdbtools_cfg)
        try:
            obj.load_data()
            asns = obj.asns
            if not asns:
                logging.warning("No origin ASNs found in "
                                "{} for {}".format(
                                    obj.descr, used_by_descr))
            return asns
        except ARouteServerError as e:
            errors = True
            logging.error(
                "Error while retrieving origin ASNs from {} for {}: {}".format(
                    obj.descr, used_by_descr, str(e)
                )
            )

        if errors:
            raise BuilderError()

class IRRDBConfigEnricher(BaseConfigEnricher):

    WORKER_THREAD_CLASS = None

    def prepare(self):
        if self.builder.irrdb_info is not None:
            return

        # In the end, self.builder.irrdb_info will be a dict
        # with this structure:
        # {
        #   "<as_set_bundle_id>": <AS_SET_Bundle_Proxy>
        # }

        irrdb_info = []

        def use_as_set(as_set_names, used_by_client=None):
            # New bundle for the given AS-SET names.
            # If a bundle for the same AS-SET names is already
            # present in the irrdb_info list, then that one will
            # be returned; otherwise, new_bundle will be.
            new_bundle = AS_SET_Bundle_Proxy(as_set_names)

            for as_set_bundle in irrdb_info:
                if as_set_bundle.id == new_bundle.id:
                    # A bundle for the same AS-SET names has been already
                    # added to irrdb_info: use it.
                    if used_by_client:
                        as_set_bundle.used_by.append(used_by_client)
                    return as_set_bundle.id

            if used_by_client:
                new_bundle.used_by.append(used_by_client)

            irrdb_info.append(new_bundle)
            return new_bundle.id

        # Add to irrdb_info all the AS-SET bundles reported in the 'asns' section.
        for asn in self.builder.cfg_asns.cfg["asns"]:
            self.builder.cfg_asns[asn]["as_set_bundle_ids"] = []

            if not self.builder.cfg_asns[asn]["as_sets"]:
                continue

            self.builder.cfg_asns[asn]["as_set_bundle_ids"].append(
                use_as_set(self.builder.cfg_asns[asn]["as_sets"])
            )

        # Add to irrdb_info all the AS-SET bundles reported in the 'clients' section.
        for client in self.builder.cfg_clients.cfg["clients"]:
            client_irrdb = client["cfg"]["filtering"]["irrdb"]
            client_irrdb["as_set_bundle_ids"] = []

            if not client_irrdb["enforce_origin_in_as_set"] and \
                not client_irrdb["enforce_prefix_in_as_set"] and \
                not self.builder.cfg_general["filtering"]["irrdb"]["tag_as_set"]:
                    # Client does not require AS-SETs info to be gathered.
                    continue

            if self.builder.ip_ver is not None:
                ip = client["ip"]
                if IPAddress(ip).version != self.builder.ip_ver:
                    # The address family of this client is not the
                    # current one used to build the configuration.
                    continue

            # Client needs AS-SETs info because origin ASN or prefix filters
            # are required.

            # In the worst case, use AS<asn>.
            client_irrdb["as_set_bundle_ids"].append(
                use_as_set(["AS{}".format(client["asn"])],
                           "client {}".format(client["id"]))
            )

            if client_irrdb["as_sets"]:
                # Client has its own specific set of AS-SETs.
                client_irrdb["as_set_bundle_ids"].append(
                    use_as_set(client_irrdb["as_sets"], "client {}".format(client["id"]))
                )
                continue

            # Client needs AS-SETs info but has not its own set of AS-SETs.

            # If client's ASN is configured in the 'asns' section and has
            # one or more AS-SETs, the client configuration will be based
            # on those.
            asn = "AS{}".format(client["asn"])
            if asn in self.builder.cfg_asns.cfg["asns"] and \
                self.builder.cfg_asns.cfg["asns"][asn]["as_sets"]:
                client_irrdb["as_set_bundle_ids"].append(
                    use_as_set(
                        self.builder.cfg_asns.cfg["asns"][asn]["as_sets"],
                        "client {}".format(client["id"])
                    )
                )
                continue

            # If one or more AS-SETs have been found on PeeringDB,
            # use them.
            as_sets_from_pdb = client.get("as_sets_from_pdb", None)
            if as_sets_from_pdb:
                logging.info("No AS-SETs provided for the '{}' client. "
                             "Using AS{} + those obtained from PeeringDB: "
                             "{}.".format(
                                    client["id"], client["asn"],
                                    ", ".join(as_sets_from_pdb)
                                ))
                client_irrdb["as_set_bundle_ids"].append(
                    use_as_set(as_sets_from_pdb, "client {}".format(client["id"]))
                )
                continue

            # No other AS-SETs found for the client.
            logging.warning("No AS-SETs provided for the '{}' client. "
                            "Only AS{} will be expanded.".format(
                                client["id"], client["asn"]
                            ))

        # Removing unreferenced AS-SETs.
        for as_set_bundle in irrdb_info:
            if not as_set_bundle.used_by:
                logging.debug("Removing unreferenced AS-SET: "
                              "{}".format(as_set_bundle.name))
        irrdb_info = [bundle for bundle in irrdb_info if bundle.used_by]

        self.builder.irrdb_info = {}
        for as_set_bundle in irrdb_info:
            self.builder.irrdb_info[as_set_bundle.id] = as_set_bundle

    def _config_thread(self, thread):
        thread.ip_ver = self.builder.ip_ver
        thread.irrdbtools_cfg = {
            "bgpq3_path": self.builder.bgpq3_path,
            "bgpq3_host": self.builder.bgpq3_host,
            "bgpq3_sources": self.builder.bgpq3_sources,
            "cache_dir": self.builder.cache_dir,
            "cache_expiry": self.builder.cache_expiry,
        }

    def add_tasks(self):
        # Enqueuing tasks.
        for as_set_bundle_id, as_set_bundle in iteritems(self.builder.irrdb_info):
            self.tasks_q.put(as_set_bundle)

class IRRDBConfigEnricher_OriginASNs(IRRDBConfigEnricher):

    WORKER_THREAD_CLASS = IRRDBConfigEnricher_WorkerThread_OriginASNs

class IRRDBConfigEnricher_Prefixes(IRRDBConfigEnricher):

    WORKER_THREAD_CLASS = IRRDBConfigEnricher_WorkerThread_Prefixes

    def _config_thread(self, thread):
        IRRDBConfigEnricher._config_thread(self, thread)
        thread.irrdbtools_cfg["allow_longer_prefixes"] = \
            self.builder.cfg_general["filtering"]["irrdb"]["allow_longer_prefixes"]



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six

from pierky.arouteserver.tests.live_tests.base import LiveScenario
from pierky.arouteserver.tests.live_tests.bird import BIRDInstanceIPv4, \
                                                      BIRDInstanceIPv6
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPD60Instance, \
                                                          OpenBGPD61Instance, \
                                                          OpenBGPD62Instance

# -----------------------------------------------------------------------
# FULL DOCUMENTATION ON
#
# https://arouteserver.readthedocs.io/en/latest/LIVETESTS.html#how-to-build-custom-scenarios
# -----------------------------------------------------------------------

class SkeletonScenario(LiveScenario):
    """The base class that describes your scenario.

    A multi level structure of parent/child classes allows to decuple
    test functions (that is, the scenario expectations) from the BGP
    speakers configuration and IP addresses and prefixes that are used
    to build the scenario.

    The example structure looks like the following:

    - base.py: a base class (this one) where test functions are implemented
      in an IP version independent way;

    - test_XXX.py: BGP speaker specific and IP version specific
      classes where the real IP addresses and prefixes are provided
      in a dictionary made of ``prefix_ID: real_IP_prefix`` entries.

    If it's needed by the scenario, the derived classes must also fill the
    ``AS_SET`` and ``R_SET`` dictionaries with the expected content of any
    expanded AS-SETs used in IRRDB validation:

    - ``AS_SET``'s items must be in the format
      ``<AS_SET_name>: <list_of_authorized_origin_ASNs>``.

    - ``R_SET``'s items must be in the format
      ``<AS_SET_name>: <list_of_authorized_prefix_IDs>`` (where prefix
      IDs are those reported in the ``DATA`` dictionary).

    Example::

      AS_SET = {
          "AS-AS1": [1],
          "AS-AS1_CUSTOMERS": [101],
          "AS-AS2": [2],
          "AS-AS2_CUSTOMERS": [101]
      }
      R_SET = {
          "AS-AS1": [
              "AS1_allowed_prefixes"
          ],
          "AS-AS1_CUSTOMERS": [
              "AS101_prefixes"
          ]
      }

    Finally, this class must implement all the tests that are shared between
    the IPv4 and the IPv6 version of this scenario.

    Writing test functions
    ----------------------

    Test functions names must start with "test_"; tests are processed in
    alphabetical order; each test is independent from the others.

    Some helper functions can be used to define expectations.

    - ``self.session_is_up()``: test if a BGP session between the two
        instances is up.

        Details here (URL wraps):

        http://arouteserver.readthedocs.io/en/latest/LIVETESTS_CODEDOC.html#
        pierky.arouteserver.tests.live_tests.base.LiveScenario.session_is_up

    - ``self.receive_route()``: test if the BGP speaker receives the expected
        route(s).

        Details here (URL wraps):

        http://arouteserver.readthedocs.io/en/latest/LIVETESTS_CODEDOC.html#
        pierky.arouteserver.tests.live_tests.base.LiveScenario.receive_route


    - ``self.log_contains()``: test if the BGP speaker's log contains the
        expected message.

        Details here (URL wraps):

        http://arouteserver.readthedocs.io/en/latest/LIVETESTS_CODEDOC.html#
        pierky.arouteserver.tests.live_tests.base.LiveScenario.log_contains
    """

    # Leave this to False to avoid nose to use this abstract class to run
    # tests. Only derived, more specific classes (test_XXX.py) must have
    # this set to True.
    __test__ = False

    # This allows to use files and directories paths which are relative
    # to this scenario root directory.
    MODULE_PATH = __file__

    # The following attributes must be setted in derived classes.
    CONFIG_BUILDER_CLASS = None
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    IP_VER = None

    # If needed for IRRDB validation, fill this dictionary with pairs
    # in the format "<AS_SET_name>": [<list_of_authorized_origin_ASNs>].
    # See the example in the class docstring above.
    AS_SET = {
    }

    # If needed for IRRDB validation, fill this dictionary with pairs
    # in the format "<AS_SET_name>": [<list_of_authorized_prefix_IDs>].
    # See the example in the class docstring above.
    R_SET = {
    }

    @classmethod
    def _setup_instances(cls):
        """Declare the BGP speaker instances that are used in this scenario.

        The ``cls.INSTANCES`` attribute is a list of all the instances that
        are used in this scenario. It is used to render local Jinja2 templates
        and to transform them into real BGP speaker configuration files.

        The ``cls.RS_INSTANCE_CLASS`` and ``cls.CLIENT_INSTANCE_CLASS``
        attributes are set by the derived classes (test_XXX.py) and
        represent the route server class and the other BGP speakers class
        respectively.

        - The first argument is the instance name.

        - The second argument is the IP address that is used to run the
          instance. Here, the ``cls.DATA`` dictionary is used to lookup the
          real IP address to use, which is configured in the derived classes
          (test_XXX.py).

        - The third argument is a list of files that are mounted from the local
          host (where Docker is running) to the container (the BGP speaker).
          The list is made of pairs in the form
          ``(local_file, container_file)``.
          The ``cls.build_rs_cfg`` and ``cls.build_other_cfg`` helper functions
          allow to render Jinja2 templates and to obtain the path of the local
          output files.

          For the route server, the configuration is built using ARouteServer's
          library on the basis of the options given in the YAML files.

          For the other BGP speakers, the configuration must be provided in the
          Jinja2 files within the scenario directory.
        """

        cls.INSTANCES = [
            cls._setup_rs_instance(),

            cls.CLIENT_INSTANCE_CLASS(
                "AS1",
                cls.DATA["AS1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            )
        ]

    @classmethod
    def _setup_rs_instance(cls):
        if cls.RS_INSTANCE_CLASS is OpenBGPD60Instance:
            return cls.RS_INSTANCE_CLASS(
                "rs",
                cls.DATA["rs_IPAddress"],
                [
                    (
                        cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                         target_version="6.0"),
                        "/etc/bgpd.conf"
                    )
                ]
            )
        if cls.RS_INSTANCE_CLASS is OpenBGPD61Instance:
            return cls.RS_INSTANCE_CLASS(
                "rs",
                cls.DATA["rs_IPAddress"],
                [
                    (
                        cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                         target_version="6.1"),
                        "/etc/bgpd.conf"
                    )
                ]
            )
        if cls.RS_INSTANCE_CLASS is OpenBGPD62Instance:
            return cls.RS_INSTANCE_CLASS(
                "rs",
                cls.DATA["rs_IPAddress"],
                [
                    (
                        cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                         target_version="6.2"),
                        "/etc/bgpd.conf"
                    )
                ]
            )
        if cls.RS_INSTANCE_CLASS is BIRDInstanceIPv4 or \
            cls.RS_INSTANCE_CLASS is BIRDInstanceIPv6:
            return cls.RS_INSTANCE_CLASS(
                "rs",
                cls.DATA["rs_IPAddress"],
                [
                    (
                        cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER),
                        "/etc/bird/bird.conf"
                    )
                ]
            )
        raise NotImplementedError("RS_INSTANCE_CLASS unknown: {}".format(
            cls.RS_INSTANCE_CLASS.__name__))

    def set_instance_variables(self):
        """Simply set local attributes for an easier usage later

        The argument of ``self._get_instance_by_name()`` must be one of
        the instance names used in ``_setup_instances()``.
        """
        self.AS1 = self._get_instance_by_name("AS1")
        self.AS2 = self._get_instance_by_name("AS2")
        self.rs = self._get_instance_by_name("rs")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1)
        self.session_is_up(self.rs, self.AS2)

    def test_030_rs_receives_AS2_prefix(self):
        """{}: rs receives AS2 prefix"""
        self.receive_route(self.rs, self.DATA["AS2_prefix1"],
                           other_inst=self.AS2, as_path="2")

    def test_030_rs_rejects_bogon(self):
        """{}: rs rejects bogon prefix"""
        self.log_contains(self.rs,
                          "prefix is bogon - REJECTING {}".format(
                              self.DATA["AS2_bogon1"]))
        self.receive_route(self.rs, self.DATA["AS2_bogon1"],
                           other_inst=self.AS2, as_path="2",
                           filtered=True)
        # AS1 should not receive the bogon prefix from the route server
        with six.assertRaisesRegex(self, AssertionError, "Routes not found"):
            self.receive_route(self.AS1, self.DATA["AS2_bogon1"])

    def test_030_custom_test(self):
        """{}: custom test"""



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import ConfigParserBase
from .validators import *
from ..errors import ConfigError, ARouteServerError


class ConfigParserASNS(ConfigParserBase):

    ROOT = "asns"

    def parse(self):
        if "clients" in self.cfg:
            del self.cfg["clients"]
        if "asns" not in self.cfg:
            self.cfg["asns"] = {}
            return

        errors = False

        schema = {
            "as_sets": ValidatorListOf(ValidatorASSet,
                                       mandatory=False)
        }

        for asn in self.cfg["asns"]:
            try:
                if not asn.startswith("AS"):
                    raise ConfigError()
                if not ValidatorASN().validate(asn[2:]):
                    raise ConfigError()
            except ConfigError:
                logging.error(
                    "Invalid ASN format in 'asns' section for '{}': "
                    "it must be in the 'AS<asn>' format.".format(asn)
                )
                errors = True

            try:
                ConfigParserBase.validate(schema, self.cfg["asns"][asn], "asns")
            except ARouteServerError as e:
                err_msg = ("One or more errors occurred while processing "
                           "the 'asns' configuration for '{}'".format(asn))
                if str(e):
                    err_msg += ": " + str(e)
                logging.error(err_msg)
                errors = True

        if errors:
            raise ConfigError()



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging

from .base import ConfigParserBase
from .validators import ValidatorPrefixListEntry
from ..errors import ConfigError, ARouteServerError


class ConfigParserBogons(ConfigParserBase):

    ROOT = "bogons"

    def parse(self):
        """
        Contents of cfg dict is updated/normalized by validators.
        """

        if "bogons" not in self.cfg:
            raise ConfigError("Missing 'bogons' top element.")

        errors = False

        for bogon in self.cfg["bogons"]:
            try:
                bogon = ValidatorPrefixListEntry().validate(bogon)
            except ARouteServerError as e:
                msg = "Error in bogon definition"
                if str(e):
                    msg += ": {}.".format(str(e))
                logging.error(msg)
                errors = True

        if errors:
            raise ConfigError()



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import yaml


from .validators import ConfigParserValidator
from ..errors import ConfigError, MissingFileError, ARouteServerError


class ConfigParserBase(object):

    ROOT = None

    def __init__(self):
        self.cfg = None
        self.file_dir = None

    def __contains__(self, name):
        return name in self.cfg[self.ROOT]

    def __getitem__(self, name):
        return self.cfg[self.ROOT][name]

    def __setitem__(self, name, val):
        self.cfg[self.ROOT][name] = val

    def __delitem__(self, name):
        del self.cfg[self.ROOT][name]

    def _load_from_yaml(self, doc):

        def expand_include(lines):
            res = ""
            for line in lines:
                if not line:
                    continue

                if not line.strip():
                    continue

                if res:
                    res += "\n"

                if line.strip().startswith("!include"):
                    filepath = line.strip().split(" ")[1]
                    filepath = os.path.expanduser(filepath)

                    if not os.path.isabs(filepath):
                        if self.file_dir:
                            filepath = os.path.join(self.file_dir, filepath)

                    with open(filepath) as inputfile:
                        raw = inputfile.read()

                    res += expand_include(raw.split("\n"))
                    continue

                res += line

            return res

        lines = doc.split("\n")
        expanded_doc = expand_include(lines)

        try:
            self.cfg = yaml.safe_load(expanded_doc)
        except Exception as e:
            raise ConfigError(
                "Can't parse YAML file: {}".format(str(e))
            )

    def _load_from_yaml_file(self, cfg_path):
        if not os.path.isfile(cfg_path):
            raise MissingFileError(cfg_path)

        self.file_dir = os.path.dirname(cfg_path)

        with open(cfg_path, "r") as f:
            self._load_from_yaml(f.read())

    def load(self, cfg_path):
        self._load_from_yaml_file(cfg_path)

        try:
            self.parse()
        except ARouteServerError as e:
            if str(e):
                logging.error(str(e))
            raise ConfigError()

    @staticmethod
    def validate(schema, cfg, path=""):
        errors = False
        if cfg is None:
            cfg = {}
            return

        for prop in cfg:
            if prop not in schema:
                errors = True
                logging.error(
                    "Unknown statement at '{}' level: '{}'.".format(
                        path, prop
                    )
                )

        for prop in schema:
            if isinstance(schema[prop], ConfigParserValidator):
                validator = schema[prop]

                try:
                    if prop in cfg:
                        cfg[prop] = validator.validate(cfg[prop])
                    else:
                        cfg[prop] = validator.validate(None)
                except ConfigError as e:
                    errors = True
                    logging.error(
                        "Error parsing '{}' at '{}' level - {}.".format(
                            prop, path, str(e)
                        )
                    )

            elif isinstance(schema[prop], dict):
                if prop not in cfg:
                    cfg[prop] = {}
                if cfg[prop] is None:
                    cfg[prop] = {}

                ConfigParserBase.validate(
                    schema[prop], cfg[prop],
                    prop if path == "" else "{}.{}".format(path, prop)
                )

            else:
                raise NotImplementedError()

        if errors:
            raise ConfigError()

    def parse(self):
        """
        Contents of cfg dict is updated/normalized by validators.
        """
        raise NotImplementedError()

def convert_next_hop_policy(cfg):
    if not cfg:
        return
    if not isinstance(cfg, dict):
        return
    if "filtering" not in cfg:
        return
    if not cfg["filtering"]:
        return
    if not isinstance(cfg["filtering"], dict):
        return
    if "next_hop_policy" in cfg["filtering"]:
        if "next_hop" in cfg["filtering"]:
            raise ConfigError(
                "Can't use the new 'next_hop' and the old 'next_hop_policy' "
                "statements simultaneously for NEXT_HOP policy configuration"
            )

        cfg["filtering"]["next_hop"] = {
            "policy": cfg["filtering"]["next_hop_policy"]
        }
        del cfg["filtering"]["next_hop_policy"]
